/*
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to breakpoint().  Breakpoint()
 *  simulates a breakpoint by executing a trap #1.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 *    bBB..BB	    Set baud rate to BB..BB		   OK or BNN, then sets
 *							   baud rate
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <node.h>
#include <asm.h>
#include <conc.h>

#ifdef DEBUG
#define debug_msg fprintf
#else
debug_msg() {}
#endif

/************************************************************************
 *
 * external low-level support routines
 */

extern putDebugChar();   /* write a single character      */
extern getDebugChar();   /* read and return a single char */

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
#define BUFMAX 2048

static int initialized = 0;	/* !0 means we've been initialized */

static const char hexchars[]="0123456789abcdef";

#define NUMREGS 8

/* Number of bytes of registers.  */
#define NUMREGBYTES (NUMREGS * 4)
enum regnames { IPTR, WPTR, AREG, BREG, CREG, FAREG, FBREG, FCREG };

#define BRKMAX	10
struct {
	int InUse;
	unsigned long Address;
	unsigned char Contents;
}BreakStruct[BRKMAX];
#define J_ZERO	0

unsigned long registers[NUMREGS];
/* Convert ch from a hex digit to an int */

static int
hex(ch)
     unsigned char ch;
{
  if (ch >= 'a' && ch <= 'f')
    return ch-'a'+10;
  if (ch >= '0' && ch <= '9')
    return ch-'0';
  if (ch >= 'A' && ch <= 'F')
    return ch-'A'+10;
  return -1;
}

#ifdef GDB_SLOW
getpacket(buffer)
     char *buffer;
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int i;
  int count;
  unsigned char ch;

  do
    {
      /* wait around for the start character, ignore all other characters */
      while ((ch = getDebugChar()) != '$') ;

      checksum = 0;
      xmitcsum = -1;

      count = 0;

      /* now, read until a # or end of buffer is found */
      while (count < BUFMAX)
	{
	  ch = getDebugChar();
	  if (ch == '#')
	    break;
	  checksum = checksum + ch;
	  buffer[count] = ch;
	  count = count + 1;
	}

      if (count >= BUFMAX)
	continue;

      buffer[count] = 0;

      if (ch == '#')
	{
	  xmitcsum = hex(getDebugChar()) << 4;
	  xmitcsum |= hex(getDebugChar());
	  if (checksum != xmitcsum)
	    putDebugChar('-');	/* failed checksum */
	  else
	    {
	      putDebugChar('+'); /* successful transfer */
	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  putDebugChar(buffer[0]);
		  putDebugChar(buffer[1]);
		  /* remove sequence chars from buffer */
		  count = strlen(buffer);
		  for (i=3; i <= count; i++)
		    buffer[i-3] = buffer[i];
		}
	    }
	}
    }
  while (checksum != xmitcsum);
}

/* send the packet in buffer.  */

void
putpacket(buffer)
     unsigned char *buffer;
{
  unsigned char checksum;
  int count;
  unsigned char ch;

  /*  $<packet info>#<checksum>. */
  do
    {
      putDebugChar('$');
      checksum = 0;
      count = 0;

      while (ch = buffer[count])
	{
	  if (! putDebugChar(ch))
	    return;
	  checksum += ch;
	  count += 1;
	}

      putDebugChar('#');
      putDebugChar(hexchars[checksum >> 4]);
      putDebugChar(hexchars[checksum & 0xf]);

    }
  while (getDebugChar() != '+');
}
#else
void
getpacket(buffer)
     char *buffer;
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int i;
  int count, len;
  unsigned char ch;
  unsigned char tempBuf[3];

  do
    {
#if 0
      /* wait around for the start character, ignore all other characters */
      while ((ch = getDebugChar()) != '$') ;

      checksum = 0;
      xmitcsum = -1;

      getDebugBuf( tempBuf, 2 );
      count = hex(tempBuf[0]) << 4;
      count |= hex(tempBuf[1]);
      count = getDebugBuf( buffer, count );
#else
      do {
         len = getDebugBuf( tempBuf, 3 );
      } while (( tempBuf[0] != '$' )&&( tempBuf[1] != '$' )&&( tempBuf[2] != '$' ));
      if ( tempBuf[1] == '$' ){
        tempBuf[1] = tempBuf[2];
        tempBuf[2] = getDebugChar();
      }
      else if ( tempBuf[2] == '$' ){
	      getDebugBuf( tempBuf+1, 2 );
      }
      count = hex(tempBuf[1]) << 4;
      count |= hex(tempBuf[2]);
      len = count+3;	/* <packet>#<checksum> */
      checksum = 0;
      xmitcsum = -1;
      len = getDebugBuf( buffer, len );
#endif

      /* now, read until a # or end of buffer is found */
      for ( i = 0; i < len; i++ )
	{
          ch = buffer[i];
	  if (ch == '#')
	    break;
	  checksum = checksum + ch;
	}

      if (i > count)
	continue;

      buffer[count] = 0;

      if (ch == '#')
	{
	  xmitcsum = hex(buffer[i+1]) << 4;
	  xmitcsum |= hex(buffer[i+2]);
	  if (checksum != xmitcsum)
	    putDebugChar('-');	/* failed checksum */
	  else
	    {
	      tempBuf[0] = '+'; /* successful transfer */
	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  tempBuf[1] = buffer[0];
		  tempBuf[2] = buffer[1];
                  putDebugBuf( tempBuf, 3 );
		  /* remove sequence chars from buffer */
		  count = strlen(buffer);
		  for (i=3; i <= count; i++)
		    buffer[i-3] = buffer[i];
		}
              else
                putDebugChar( '+' );
	    }
	}
    }
  while (checksum != xmitcsum);
}

/* send the packet in buffer.  */

void
putpacket(buffer)
     unsigned char *buffer;
{
  unsigned char checksum;
  int count, len;
  unsigned char ch;
  unsigned char *tempBuf;

  checksum = 0;
  count = 0;
  while (ch = buffer[count])
	{
	  checksum += ch;
	  count += 1;
	}
  count += 4;
  tempBuf = (unsigned char *)malloc( count * sizeof(unsigned char));
  tempBuf[0] = '$';
  count = 0;
  while (ch = buffer[count])
	{
          tempBuf[count+1] = ch;
	  count += 1;
	}
  count++;
  tempBuf[count++] = '#';
  tempBuf[count++] = hexchars[checksum >> 4];
  tempBuf[count++] = hexchars[checksum & 0xf];
  /*  $<packet info>#<checksum>. */
  do
    {
       putDebugBuf( tempBuf, count );
    }
  while (getDebugChar() != '+');
}
#endif

static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/* Indicate to caller of mem2hex or hex2mem that there has been an
   error.  */
static volatile int mem_err = 0;

/* Convert the memory pointed to by mem into hex, placing result in buf.
 * Return a pointer to the last char put in buf (null), in case of mem fault,
 * return 0.
 * If MAY_FAULT is non-zero, then we will handle memory faults by returning
 * a 0, else treat a fault like any other fault in the stub.
 */

#if 1
/*  Conversion done by GDB */
static unsigned char *
mem2hex(mem, buf, count, may_fault)
     unsigned char *mem;
     unsigned char *buf;
     int count;
     int may_fault;
{
  unsigned char ch;

  while (count-- > 0)
    {
      ch = *mem++;
      if (mem_err)
	return 0;
      *buf++ = hexchars[ch >> 4];
      *buf++ = hexchars[ch & 0xf];
    }

  *buf = 0;

  return buf;
}
#else
static unsigned char *
mem2hex(mem, buf, count, may_fault)
     unsigned char *mem;
     unsigned char *buf;
     int count;
     int may_fault;
{
  unsigned char ch;
  int index, i;

  index = 0;
  while (1)
    {
      if ( count == 0 ) break;
      if ( count >= 4 ){
        index += 3;
        for ( i = 0; i < 4; i++ ){
          ch = mem[index]; index--;
          if (mem_err)
	    return 0;
          *buf++ = hexchars[ch >> 4];
          *buf++ = hexchars[ch & 0xf];
        }
        index += 4;
        count -= 4;
      }
      else if ( count >= 2 ){
        index += 1;
        for ( i = 0; i < 2; i++ ){
          ch = mem[index]; index--;
          if (mem_err)
	    return 0;
          *buf++ = hexchars[ch >> 4];
          *buf++ = hexchars[ch & 0xf];
        }
        index += 2;
        count -= 2;
      }
      else {
          ch = mem[index]; 
          if (mem_err)
	    return 0;
          *buf++ = hexchars[ch >> 4];
          *buf++ = hexchars[ch & 0xf];
          index++;
          count--;
      }
    }

  *buf = 0;

  return buf;
}
#endif

/* convert the hex array pointed to by buf into binary to be placed in mem
 * return a pointer to the character AFTER the last byte written */

static char *
hex2mem(buf, mem, count, may_fault)
     unsigned char *buf;
     unsigned char *mem;
     int count;
     int may_fault;
{
  int i;
  unsigned char ch;

  for (i=0; i<count; i++)
    {
      ch = hex(*buf++) << 4;
      ch |= hex(*buf++);
      *mem++ = ch;
      if (mem_err)
	return 0;
    }

  return mem;
}


/* Set up exception handlers for tracing and breakpoints */

int TrapWptr=0, TrapIptr = 0;
extern void TrapHandlerThread();

void
set_debug_traps()
{
  unsigned long *memstart;
  Process *pid;
  memstart = (unsigned long *)_Node.memstart;

  pid = ProcAlloc( TrapHandlerThread, 1024, 0 );
  ProcRunHigh(pid);

/*
  _ProcToHigh();
*/
  ((void * (*) (const char *)) __asm)( "setj0break" );
  memstart[0] = memstart[2] = (unsigned long)TrapWptr;
  memstart[1] = memstart[3] = (unsigned long )TrapIptr;
/*
  _ProcToLow();
*/
  debug_msg( stderr, "MemStart %x Wptr %x Iptr %x\n", memstart, TrapWptr, TrapIptr );
  
  /* In case GDB is started before us, ack any packets (presumably
     "$?#xx") sitting there.  */

  putDebugChar ('+');

  initialized = 1;
}


/* Convert the SPARC hardware trap type code to a unix signal number. */

static int
computeSignal(tt)
     int tt;
{
  return SIGTRAP;
}

/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */

static int
hexToInt(char **ptr, int *intValue)
{
  int numChars = 0;
  int hexValue;

  *intValue = 0;

  while (**ptr)
    {
      hexValue = hex(**ptr);
      if (hexValue < 0)
	break;

      *intValue = (*intValue << 4) | hexValue;
      numChars ++;

      (*ptr)++;
    }

  return (numChars);
}

/*
 * This function does all command procesing for interfacing to gdb.  It
 * returns 1 if you should skip the instruction at the trap address, 0
 * otherwise.
 */

extern void breakinst();

void
handle_exception (registers)
     unsigned long *registers;
{
  int tt;			/* Trap type */
  int sigval;
  int addr;
  int length;
  char *ptr;
  unsigned long *sp;
  static int Initialized = 0;
  unsigned long *memstart;
  int i;
  unsigned long iptr, wptr;
  static char first_time = 1;
  int ThreadId = 0;
  int WptrOffset;

  memstart = (unsigned long *)_Node.memstart;
  debug_msg( stderr, "handle_exception memstart %x %x %x %x %x %x\n", memstart[0], memstart[1], memstart[2], memstart[3], memstart[4], memstart[5] );
  WptrOffset = ( memstart[0] == TrapWptr )?2:0;
  if ( first_time ){
     first_time = 0;
     iptr = registers[IPTR] = memstart[WptrOffset+1];
  }
  else
    iptr = registers[IPTR] = memstart[WptrOffset+1]-1;
  wptr = registers[WPTR] = memstart[WptrOffset];
  debug_msg( stderr, "Trap Iptr %x Wptr %x\n", iptr, wptr );
  /* reply to host that an exception has occurred */
  sigval = computeSignal(tt);

  if ( !Initialized ){
    ptr = remcomOutBuffer;
    *ptr++ = 'N';
    *ptr++ = hexchars[sigval >> 4];
    *ptr++ = hexchars[sigval & 0xf];
    sprintf (remcomOutBuffer + 3, "0x%x;0x%x;0x%x",
               _Node.process->code,
               _Node.process->data,
               _Node.process->data);
    putpacket(remcomOutBuffer);
    Initialized = 1;
    debug_msg( stderr, "Relocate %s, datasize %x\n", remcomOutBuffer+3,  _Node.process->datasize );
  }

  ptr = remcomOutBuffer;
  *ptr++ = 'T';
  *ptr++ = hexchars[sigval >> 4];
  *ptr++ = hexchars[sigval & 0xf];

  *ptr++ = hexchars[ThreadId >> 4];
  *ptr++ = hexchars[ThreadId & 0xf];
  *ptr++ = ';';

  *ptr++ = hexchars[IPTR >> 4];
  *ptr++ = hexchars[IPTR & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&registers[IPTR], ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[WPTR >> 4];
  *ptr++ = hexchars[WPTR & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&registers[WPTR], ptr, 4, 0); /* WPTR */
  *ptr++ = ';';

  *ptr++ = hexchars[AREG >> 4];
  *ptr++ = hexchars[AREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&registers[AREG], ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[BREG >> 4];
  *ptr++ = hexchars[BREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&registers[BREG], ptr, 4, 0);
  *ptr++ = ';';

  *ptr++ = hexchars[CREG >> 4];
  *ptr++ = hexchars[CREG & 0xf];
  *ptr++ = ':';
  ptr = mem2hex((char *)&registers[CREG], ptr, 4, 0);
  *ptr++ = ';';


  *ptr++ = 0;

  putpacket(remcomOutBuffer);

  while (1)
    {
      remcomOutBuffer[0] = 0;

      getpacket(remcomInBuffer);
      debug_msg( stderr, "Received %s\n", remcomInBuffer );
      switch (remcomInBuffer[0])
	{
	case '?':
#if 0
	  remcomOutBuffer[0] = 'S';
	  remcomOutBuffer[1] = hexchars[sigval >> 4];
	  remcomOutBuffer[2] = hexchars[sigval & 0xf];
	  remcomOutBuffer[3] = 0;
#else
	  ptr = remcomOutBuffer;
	  *ptr++ = 'N';
	  *ptr++ = hexchars[sigval >> 4];
	  *ptr++ = hexchars[sigval & 0xf];
	  sprintf (remcomOutBuffer + 3, "0x%x;0x%x;0x%x",
		       _Node.process->code,
		       _Node.process->data,
		       _Node.process->data);
#endif
	  break;

	case 'd':
				/* toggle debug flag */
	  break;

	case 'g':		/* return the value of the CPU registers */
	  {
	    ptr = remcomOutBuffer;
	    ptr = mem2hex((char *)registers, ptr, NUMREGS * 4, 0); 
	  }
	  break;

	case 'G':	   /* set the value of the CPU registers - return OK */
	  {
	    unsigned long *newsp, psr;

	    sp = registers[WPTR]; 

	    ptr = &remcomInBuffer[1];
	    hex2mem(ptr, (char *)registers, NUMREGS * 4, 0); 

	    /* See if the stack pointer has moved.  If so, then copy the saved
	       locals and ins to the new location.  This keeps the window
	       overflow and underflow routines happy.  */

	    newsp = (unsigned long *)registers[WPTR];
	    if (sp != newsp)
#if 0
	      sp = memcpy(newsp, sp, 16 * 4);
#else
            {
	      strcpy(remcomOutBuffer,"OK");
              break;
            }
#endif


	    strcpy(remcomOutBuffer,"OK");
	  }
	  break;

	case 'm':	  /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	  /* Try to read %x,%x.  */

	  ptr = &remcomInBuffer[1];

	  if (hexToInt(&ptr, &addr)
	      && *ptr++ == ','
	      && hexToInt(&ptr, &length))
	    {
              if (( addr >= 0x80000000 )&&( addr <= 0x80400000 ))
              {
	        if (mem2hex((char *)addr, remcomOutBuffer, length, 1))
		  break;
              }

	      strcpy (remcomOutBuffer, "E03");
	    }
	  else
	    strcpy(remcomOutBuffer,"E01");
	  break;

	case 'M': /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
	  /* Try to read '%x,%x:'.  */

	  ptr = &remcomInBuffer[1];

	  if (hexToInt(&ptr, &addr)
	      && *ptr++ == ','
	      && hexToInt(&ptr, &length)
	      && *ptr++ == ':')
	    {
	      if (hex2mem(ptr, (char *)addr, length, 1))
		strcpy(remcomOutBuffer, "OK");
	      else
		strcpy(remcomOutBuffer, "E03");
	    }
	  else
	    strcpy(remcomOutBuffer, "E02");
	  break;

	case 'c':    /* cAA..AA    Continue at address AA..AA(optional) */
	  /* try to read optional parameter, pc unchanged if no parm */

	  ptr = &remcomInBuffer[1];
	  if (hexToInt(&ptr, &addr))
	    {
	      registers[IPTR] = addr;
	    }

/* Need to flush the instruction cache here, as we may have deposited a
   breakpoint, and the icache probably has no way of knowing that a data ref to
   some location may have changed something that is in the instruction cache.
 */

	  flush_i_cache();
	  debug_msg( stderr, "Iptr %x Wptr %x\n", iptr, wptr );
	  memstart[WptrOffset+1] = iptr ;
	  memstart[WptrOffset] = wptr ;
	  return;

	  /* kill the program */
	case 'k' :
	  break;
	case 'B' :
	  ptr = &remcomInBuffer[1];
	  hexToInt(&ptr, &addr);

	  for ( i = 0; i < BRKMAX; i++ ){
		if (!BreakStruct[i].InUse ){
		   BreakStruct[i].Address = addr;
		   BreakStruct[i].Contents = *(unsigned char *)addr;
		   BreakStruct[i].InUse = 1;
		   *(unsigned char *)addr = J_ZERO;
	        }
	  }
	  strcpy(remcomOutBuffer, "OK");
	  break;
	case 'b' :
	  ptr = &remcomInBuffer[1];
	  hexToInt(&ptr, &addr);
	  for ( i = 0; i < BRKMAX; i++ ){
		if (BreakStruct[i].InUse ){
		   if ( BreakStruct[i].Address == addr ){
		       *(unsigned char *)addr = BreakStruct[i].Contents;
		       BreakStruct[i].InUse = 0;
		       break;
		   }
	        }
	  }
	  if ( i < BRKMAX )
	    strcpy(remcomOutBuffer, "OK");
	  else
	    strcpy(remcomOutBuffer, "E04");
	  break;

#if 0
Disabled until we can unscrew this properly

	case 'b':	  /* bBB...  Set baud rate to BB... */
	  {
	    int baudrate;
	    extern void set_timer_3();

	    ptr = &remcomInBuffer[1];
	    if (!hexToInt(&ptr, &baudrate))
	      {
		strcpy(remcomOutBuffer,"B01");
		break;
	      }

	    /* Convert baud rate to uart clock divider */
	    switch (baudrate)
	      {
	      case 38400:
		baudrate = 16;
		break;
	      case 19200:
		baudrate = 33;
		break;
	      case 9600:
		baudrate = 65;
		break;
	      default:
		strcpy(remcomOutBuffer,"B02");
		goto x1;
	      }

	    putpacket("OK");	/* Ack before changing speed */
	    set_timer_3(baudrate); /* Set it */
	  }
x1:	  break;
#endif
	}			/* switch */

      /* reply to the request */
      putpacket(remcomOutBuffer);
    }
}

/* This function will generate a breakpoint exception.  It is used at the
   beginning of a program to sync up with a debugger and can be used
   otherwise as a quick means to stop program execution and "break" into
   the debugger. */

void
breakpoint()
{
  int temp;
  if (!initialized)
    return;

  __asm ( "j 0" );
  debug_msg( stderr, "Returned from breakpoint %x %x\n", *(int *)(0x80000078), *(int *)(0x8000007c) );
}

exceptionHandler()
{
}

