/* Target-dependent code for the SPARC for GDB, the GNU debugger.
   Copyright 1986, 1987, 1989, 1991, 1992, 1993 Free Software Foundation, Inc.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "defs.h"
#include "frame.h"
#include "inferior.h"
#include "obstack.h"
#include "target.h"
#include "ieee-float.h"

#include "symfile.h" /* for objfiles.h */
#include "objfiles.h" /* for find_pc_section */

#ifdef	USE_PROC_FS
#include <sys/procfs.h>
#endif

#include "gdbcore.h"

/* From infrun.c */
extern int stop_after_trap;

/* We don't store all registers immediately when requested, since they
   get sent over in large chunks anyway.  Instead, we accumulate most
   of the changes and send them over once.  "deferred_stores" keeps
   track of which sets of registers we have locally-changed copies of,
   so we only need send the groups that have changed.  */

int deferred_stores = 0;	/* Cumulates stores we want to do eventually. */

typedef enum
{
  Error, not_branch, jmp, c_jmp, ret, call, gcall, lend
} branch_type;

#define JMP	0x0
#define C_JMP	0xa
#define CALL	0x9
#define PFIX	0x2
#define NFIX	0x6
#define AJW	0xb
#define RET1	0x22
#define RET2	0xf0
#define GCALL	0xf6
#define LEND1	0x22
#define LEND2	0xf1

#define OPCODE( instn ) (( (instn) & 0xf0 ) >> 4 )
#define DATA( instn ) ( (instn) & 0x0f	)


static CORE_ADDR next_pc, target;
static int brkjmp, brknext;

/* Non-zero if we just simulated a single-step .  This is
   needed because we cannot remove the breakpoints in the inferior
   process until after the `wait' in `wait_for_inferior'.
*/

int one_stepped;

/* single_step() is called just before we want to resume the inferior,
   if we want to single-step it but there is no hardware or kernel single-step
   support .  We find all the possible targets of the
   coming instruction and breakpoint them.

   single_step is also called just after the inferior stops.  If we had
   set up a simulated single-step, we undo our damage.  */

void
single_step (ignore)
     int ignore; /* pid, but we don't need it */
{
  branch_type br, isannulled();
  CORE_ADDR pc, get_next_pc();
  unsigned char pc_instruction;

  if (!one_stepped)
    {
      brkjmp = 0; brknext = 0;
      pc = read_register (PC_REGNUM);
      next_pc = get_next_pc( pc );
      read_memory (pc, &pc_instruction,sizeof(pc_instruction));
      br = isannulled (pc_instruction, pc, &target);
      if ((br == c_jmp)||( br == jmp)||( br == ret )||(br == call)||(br == gcall)||(br == lend))
	{
	  transputer_insert_breakpoint (target);
          brkjmp = 1;
	}
      if (( br != jmp )&&( br != call )&&( br != ret )&&(br != gcall))
	{
          transputer_insert_breakpoint (next_pc);
          brknext = 1;
	}
      /* We are ready to let it go */
      one_stepped = 1;
      return;
    }
  else
    {
      /* Remove breakpoints */
      if ( brknext )
        transputer_remove_breakpoint (next_pc);
      if ( brkjmp )
        transputer_remove_breakpoint (target);

      one_stepped = 0;
    }
}
void
multi_step (ignore, EndAddress )
     int ignore; /* pid, but we don't need it */
     CORE_ADDR EndAddress;
{
  branch_type br, isannulled_in_range();
  CORE_ADDR pc, get_next_branch_pc();
  unsigned char pc_instruction;
  unsigned char *instruction_buf;

  if ( EndAddress <= 1 ){
    single_step(ignore);
    return;
  }
  if (!one_stepped)
    {
      brkjmp = 0; brknext = 0;
      pc = read_register (PC_REGNUM);
      instruction_buf = (unsigned char *)malloc( sizeof(unsigned char)* (EndAddress-pc+1));
      read_memory( pc, instruction_buf, EndAddress-pc+1);
      next_pc = get_next_branch_pc( pc, EndAddress, instruction_buf );
      br = isannulled_in_range ( pc, &target, EndAddress, instruction_buf );
      if ((br == c_jmp)||( br == jmp)||( br == ret )||(br == call)||(br == gcall)||(br == lend))
	{
	  transputer_insert_breakpoint (target);
          brkjmp = 1;
	}
      if (( br != jmp )&&( br != call )&&( br != ret )&&(br != gcall))
	{
          transputer_insert_breakpoint (next_pc);
          brknext = 1;
	}
      /* We are ready to let it go */
      one_stepped = 1;
      return;
    }
  else
    {
      /* Remove breakpoints */
      if ( brknext )
        transputer_remove_breakpoint (next_pc);
      if ( brkjmp )
        transputer_remove_breakpoint (target);

      one_stepped = 0;
    }
}


#define	FRAME_SAVED_L0	0			    /* Byte offset from SP */
#define	FRAME_SAVED_I0	(8 * REGISTER_RAW_SIZE (0)) /* Byte offset from SP */

CORE_ADDR
transputer_frame_chain (thisframe)
     FRAME thisframe;
{
  char buf[MAX_REGISTER_RAW_SIZE];
  int err;
  CORE_ADDR addr;

  addr = thisframe->frame;
  err = target_read_memory (addr, buf, sizeof(long));
  if (err)
    return 0;
  addr = extract_address (buf, sizeof(long));
  return addr;
}

CORE_ADDR
transputer_extract_struct_value_address (regbuf)
     char regbuf[REGISTER_BYTES];
{
  CORE_ADDR addr;
  addr = read_memory_integer( ((int *)(regbuf))[SP_REGNUM]+4, sizeof(long) );
  return read_memory_integer (addr-8, sizeof(long));
}

/* Find the pc saved in frame FRAME.  */

CORE_ADDR
frame_saved_pc (frame)
     FRAME frame;
{
  char buf[MAX_REGISTER_RAW_SIZE];
  CORE_ADDR addr;

  addr = (frame->frame - 20);
  read_memory (addr, buf, sizeof(long));
  return extract_address (buf, sizeof(long));
}

/*
 * Since an individual frame in the frame cache is defined by two
 * arguments (a frame pointer and a stack pointer), we need two
 * arguments to get info for an arbitrary stack frame.  This routine
 * takes two arguments and makes the cached frames look as if these
 * two arguments defined a frame on the cache.  This allows the rest
 * of info frame to extract the important arguments without
 * difficulty. 
 */
FRAME
setup_arbitrary_frame (argc, argv)
     int argc;
     FRAME_ADDR *argv;
{
  FRAME fid;

  if (argc != 1)
    error ("T800 frame specification requires one argument:  Wptr");

  fid = create_new_frame (argv[0], 0);

  if (!fid)
    fatal ("internal: create_new_frame returned invalid frame id");
  
  fid->pc = FRAME_SAVED_PC (fid);
  return fid;
}

/* Given a pc value, skip it forward past the function prologue by
   disassembling instructions that appear to be a prologue.

   If FRAMELESS_P is set, we are only testing to see if the function
   is frameless.  This allows a quicker answer.

   This routine should be more specific in its actions; making sure
   that it uses the same register in the initial prologue section.  */
CORE_ADDR 
skip_prologue (start_pc, frameless_p)
     CORE_ADDR start_pc;
     int frameless_p;
{
  return start_pc;
}

/* Check instruction at ADDR to see if it is an annulled branch.
   Set *TARGET if we find a candidate branch; set to zero if not. */
   
branch_type
isannulled (instruction, addr, target)
     unsigned char instruction;
     CORE_ADDR addr, *target;
{
  branch_type val = not_branch;
  long int offset;		/* Must be signed for sign-extend.  */
  unsigned char pc_instruction;
  
  *target = 0;
  offset = 0;
  pc_instruction = instruction;
  while (( OPCODE(pc_instruction) == PFIX || OPCODE(pc_instruction) == NFIX )){
    addr++;
    if ( OPCODE(pc_instruction) == PFIX )
	    offset = (offset + DATA(pc_instruction)) << 4;
    else if ( OPCODE(pc_instruction) == NFIX )
            offset = (~(offset + DATA(pc_instruction))) << 4;
      read_memory (addr, &pc_instruction, sizeof(pc_instruction));
  }
  if ( OPCODE(pc_instruction) == JMP )
    val = jmp;
  else if ( OPCODE(pc_instruction) == C_JMP )
    val = c_jmp;
  else if ( OPCODE(pc_instruction) == CALL )
    val = call;
  else if ((offset == 0)&&( pc_instruction == GCALL ))
    val = gcall;
  else if (((0x20+(offset>>4)) == LEND1 )&&( pc_instruction == LEND2 ))
     val = lend;
  if (((0x20+(offset>>4)) == RET1 )&&( pc_instruction == RET2 )){
     *target = read_memory_integer(read_sp(), sizeof(CORE_ADDR) );
     val = ret;
  }
  else if (( val == jmp || val == c_jmp )||( val == call )){
     offset += DATA(pc_instruction); addr++;
     *target = addr + offset;
  }
  else if (( val == gcall )||( val == lend ))
     *target = read_register( A_REGNUM );

  return val;
}
branch_type
isannulled_in_range (addr, target, EndAddress, instruction_buf)
     CORE_ADDR addr, *target, EndAddress;
     unsigned char *instruction_buf;
{
  branch_type val = not_branch;
  long int offset;		/* Must be signed for sign-extend.  */
  unsigned char pc_instruction;
  unsigned char opcode;
  int index = 0;
  int workspace_adjust = 0;
  int instn_num = 0;
  CORE_ADDR prev_addr = addr;
  
  *target = 0;
  while ( addr < EndAddress ){
    offset = 0; prev_addr = addr;
    pc_instruction = instruction_buf[index++];
    opcode = OPCODE(pc_instruction); addr++;
    while (( opcode == PFIX || opcode == NFIX )){
      if ( opcode == PFIX )
	      offset = (offset + DATA(pc_instruction)) << 4;
      else if ( opcode == NFIX )
              offset = (~(offset + DATA(pc_instruction))) << 4;
      pc_instruction = instruction_buf[index++];
      opcode = OPCODE(pc_instruction); addr++;
    }
    if ( opcode == JMP )
      val = jmp;
    else if ( opcode == C_JMP )
      val = c_jmp;
    else if ( opcode == CALL )
      val = call;
    else if (( offset == 0 )&&( pc_instruction == GCALL ))
      val = gcall;
    else if ( opcode == AJW )
      workspace_adjust = (offset+DATA(pc_instruction))*4;
    else if (((0x20+(offset>>4)) == LEND1 )&&( pc_instruction == LEND2 ))
       val = lend;
    if (((0x20+(offset>>4)) == RET1 )&&( pc_instruction == RET2 )){
       *target = read_memory_integer(read_sp()+workspace_adjust, sizeof(CORE_ADDR) );
       val = ret; break;
    }
    else if (( val == jmp || val == c_jmp )||( val == call )){
       offset += DATA(pc_instruction); 
       *target = addr + offset; break;
    }
    else if (( val == gcall )||( val == lend)){
       if ( instn_num == 0 ){ /* Current instruction is a gcall */
         *target = read_register(A_REGNUM); break;
       }
       else {
         *target = prev_addr; break;
	}
    }
    instn_num++;
  }

  return val;
}

transputer_store_struct_return( addr, sp )
 CORE_ADDR addr;
unsigned long sp;
{
  CORE_ADDR taddr;
  taddr = read_memory_integer( sp+4, sizeof(long) );
  return target_write_memory (taddr-8, (char *)&(addr), 4); 
}
/* transputer_frame_find_saved_regs ()

   Stores, into a struct frame_saved_regs,
   the addresses of the saved registers of frame described by FRAME_INFO.
*/
void
transputer_frame_find_saved_regs (fi, saved_regs_addr)
     struct frame_info *fi;
     struct frame_saved_regs *saved_regs_addr;
{
  register int regnum;
  FRAME_ADDR frame;
  FRAME fid = FRAME_INFO_ID (fi);

  frame = read_memory_integer(read_register (SP_REGNUM)+4, sizeof(long));
  if (!fid)
    fatal ("Bad frame info struct in FRAME_FIND_SAVED_REGS");

  memset (saved_regs_addr, 0, sizeof (*saved_regs_addr));

  for (regnum = A_REGNUM; regnum <= C_REGNUM; regnum++)
	saved_regs_addr->regs[regnum] = frame + (regnum - A_REGNUM) * 4 - 32;
  saved_regs_addr->regs[PC_REGNUM] = frame - 20;
  saved_regs_addr->regs[SP_REGNUM] = FRAME_FP (fi);
}

/* Push an empty stack frame, and record in it the current PC, regs, etc.

   Definitely see tm-t800.h for more doc of the frame format here.  */

void
transputer_push_dummy_frame ()
{
  CORE_ADDR sp, old_sp;
  char register_temp[0x140];

  old_sp = sp = read_register (SP_REGNUM);
  sp -= 16;
  write_register (SP_REGNUM, sp);
}

/* Discard from the stack the innermost frame, restoring all saved registers.

   Definitely see tm-t800.h for more doc of the frame format here.  */

void
transputer_pop_frame ()
{
  register FRAME frame = get_current_frame ();
  register CORE_ADDR pc;
  struct frame_saved_regs fsr;
  struct frame_info *fi;
  char raw_buffer[REGISTER_BYTES];

  fi = get_frame_info (frame);
  get_frame_saved_regs (fi, &fsr);
  if (fsr.regs[A_REGNUM])
    {
      read_memory (fsr.regs[A_REGNUM], raw_buffer, 32 * 4);
      write_register_bytes (REGISTER_BYTE (A_REGNUM), raw_buffer, 32 * 4);
    }
  if (fsr.regs[B_REGNUM])
    {
      read_memory (fsr.regs[B_REGNUM], raw_buffer, 32 * 4);
      write_register_bytes (REGISTER_BYTE (B_REGNUM), raw_buffer, 32 * 4);
    }
  if (fsr.regs[C_REGNUM])
    {
      read_memory (fsr.regs[C_REGNUM], raw_buffer, 32 * 4);
      write_register_bytes (REGISTER_BYTE (C_REGNUM), raw_buffer, 32 * 4);
    }
  flush_cached_frames ();
  set_current_frame ( create_new_frame (read_register (SP_REGNUM),
					read_pc ()));
}


transputer_pc_adjust()
{}

CORE_ADDR
get_next_pc( pc )
CORE_ADDR pc;
{
  unsigned char pc_instruction;
  CORE_ADDR addr;
  
  addr = pc;
  read_memory (addr, &pc_instruction,sizeof(pc_instruction));
  while (( OPCODE(pc_instruction) == PFIX || OPCODE(pc_instruction) == NFIX )){
      addr++;
      read_memory (addr, &pc_instruction, sizeof(pc_instruction));
  }
  return (addr+1);
}

CORE_ADDR
get_next_branch_pc( pc, EndAddress, instruction_buf )
CORE_ADDR pc, EndAddress;
unsigned char *instruction_buf;
{
  unsigned char pc_instruction;
  CORE_ADDR addr;
  unsigned char opcode, prev=0;
  int  index = 0;int offset;
  
  addr = pc;
  while ( addr < EndAddress ){
    pc_instruction = instruction_buf[index]; index++;
    opcode = OPCODE(pc_instruction); addr++;
    offset = 0;
    while (( opcode == PFIX || opcode == NFIX )){
        if ( opcode == PFIX )
	      offset = (offset + DATA(pc_instruction)) << 4;
        else if ( opcode == NFIX )
              offset = (~(offset + DATA(pc_instruction))) << 4;
        prev = pc_instruction;
        pc_instruction = instruction_buf[index]; index++;
        opcode = OPCODE(pc_instruction);  addr++;
    }
    if (( opcode == JMP )||( opcode == C_JMP )||( opcode == CALL ))
      break;
    else if (( (0x20+(offset>>4)) == RET1 )&&( pc_instruction == RET2 ))
      break;
    else if (( (0x20+(offset>>4)) == LEND1 )&&( pc_instruction == LEND2 ))
      break;
    else if ((offset == 0 )&&( pc_instruction == GCALL ))
      break;
  }
  return (addr);
}

int
frame_locals_address( fi )
     struct frame_info *fi;
{
  struct block *b;
  register struct symbol *sym;
  int j;
  int offset;

  b = get_frame_block(fi);
  offset = fi->frame-20;
  for (j = 0; j < BLOCK_NSYMS (b); j++){
    QUIT;
    sym = BLOCK_SYM (b, j);
    if ( strcmp(SYMBOL_NAME(sym), ".info_struct" ) == 0 ){
      offset -= (sym->type->length);
      offset += (sym->type->fields->bitpos / 8);
      return(offset);
    }
  }
  return(offset);
}

struct type * register_virtual_type( regno )
int regno;
{
	int fpstatus;
	if (( regno < FA_REGNUM )||( regno > FC_REGNUM )) return( builtin_type_int);
	fpstatus = read_register( FS_REGNUM );
	if (( fpstatus >> ((regno-FA_REGNUM+1)*2)) & 0x1 )
		return(builtin_type_double);
	else
		return(builtin_type_float);
}

int register_virtual_size( regno )
int regno;
{
	int fpstatus;
	if (( regno < FA_REGNUM )||( regno > FC_REGNUM )) return( sizeof(long));
	fpstatus = read_register( FS_REGNUM );
	if (( fpstatus >> ((regno-FA_REGNUM+1)*2)) & 0x1 )
		return(sizeof(double));
	else
		return(sizeof(float));
}

